# Copyright 2021 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""TraceType implementations for common Python types."""

from typing import Any, Hashable, Optional, Sequence, Type
from typing import Dict as PythonDict
from typing import Tuple as PythonTuple

from tensorflow.python.types import trace


class Generic(trace.TraceType):
  """Represents an arbitrary Python object."""

  def __init__(self, obj):
    self._object = obj
    self._object_hash = hash(obj)

  def is_subtype_of(self, other: trace.TraceType) -> bool:
    return self == other

  def most_specific_common_supertype(
      self, types: Sequence[trace.TraceType]) -> Optional["Generic"]:
    return self if all(self == other for other in types) else None

  def _placeholder_value(self) -> Any:
    return self._object

  def __eq__(self, other) -> bool:
    if not isinstance(other, trace.TraceType):
      return NotImplemented

    return isinstance(other, Generic) and self._object == other._object

  def __hash__(self) -> int:
    return self._object_hash

  def __repr__(self):
    return f"{self.__class__.__name__}(obj={self._object!r})"


class Weakref(Generic):
  """Represents weakref of an arbitrary Python object.

  When a function argument is a custom class, instead of making a copy of it
  just for the sake of function cache, a weakref is instead kept to save memory.
  """

  def _placeholder_value(self) -> Any:
    return self._object()

  def __eq__(self, other):
    if not isinstance(other, trace.TraceType):
      return NotImplemented

    if not isinstance(other, Weakref):
      return False

    if self._object() is None or other._object() is None:
      return False

    if self._object() is other._object():
      return True

    return self._object == other._object

  def __hash__(self):
    return self._object_hash


class OrderedCollection(trace.TraceType):
  """Represents an ordered collection of TraceType objects.

  Attributes:
    collection_type: Python type for the collection (list, tuple etc.)
    components: A corresponding sequence of TraceTypes to the values in the
      collection.
  """

  def __init__(self, collection_type: Type[Any],
               components: PythonTuple[trace.TraceType]):
    self.collection_type = collection_type
    self.components = components

  def _shallow_equal(self, other):
    return (isinstance(other, OrderedCollection) and
            self.collection_type == other.collection_type and
            len(self.components) == len(other.components))

  def _supertype_components(
      self, others: Sequence["OrderedCollection"]
  ) -> Optional[Sequence[trace.TraceType]]:
    """Helper that generates a list of per-component supertypes or None."""
    new_components = []
    for i, component in enumerate(self.components):
      common = component.most_specific_common_supertype(
          [other.components[i] for other in others])
      if common is None:
        return None
      else:
        new_components.append(common)
    return new_components

  def is_subtype_of(self, other: trace.TraceType) -> bool:
    """See base class."""
    if not self._shallow_equal(other):
      return False

    return all(
        self_component.is_subtype_of(other_component) for self_component,
        other_component in zip(self.components, other.components))

  def __eq__(self, other) -> bool:
    if not isinstance(other, trace.TraceType):
      return NotImplemented

    if not self._shallow_equal(other):
      return False

    return self.components == other.components

  def __hash__(self) -> int:
    return hash((self.collection_type, self.components))

  def __repr__(self):
    return (f"{self.__class__.__name__}(collection_type="
            f"{self.collection_type!r}, components={self.components!r})")


class List(OrderedCollection):
  """Represents a list of TraceType objects."""

  def __init__(self, *components: trace.TraceType):
    super().__init__(list, components)

  def most_specific_common_supertype(
      self, types: Sequence[trace.TraceType]) -> Optional["List"]:
    """See base class."""
    if not all(self._shallow_equal(other) for other in types):
      return None

    new_components = self._supertype_components(types)

    return None if new_components is None else List(*new_components)

  def _placeholder_value(self) -> Any:
    components = [
        component._placeholder_value()  # pylint: disable=protected-access
        for component in self.components
    ]
    return list(components)


class Tuple(OrderedCollection):
  """Represents a tuple of TraceType objects."""

  def __init__(self, *components: trace.TraceType):
    super().__init__(tuple, components)

  def most_specific_common_supertype(
      self, types: Sequence[trace.TraceType]) -> Optional["Tuple"]:
    """See base class."""
    if not all(self._shallow_equal(other) for other in types):
      return None

    new_components = self._supertype_components(types)

    return None if new_components is None else Tuple(*new_components)

  def _placeholder_value(self) -> Any:
    components = [
        component._placeholder_value()  # pylint: disable=protected-access
        for component in self.components
    ]
    return tuple(components)


class NamedTuple(OrderedCollection):
  """Represents a NamedTuple of TraceType objects."""

  def __init__(self, collection_type: Type[object],
               attributes: PythonTuple[trace.TraceType]):
    super().__init__(collection_type, attributes)

  def most_specific_common_supertype(
      self, types: Sequence[trace.TraceType]) -> Optional["NamedTuple"]:
    """See base class."""
    if not all(self._shallow_equal(other) for other in types):
      return None

    new_components = self._supertype_components(types)

    return None if new_components is None else type(self)(self.collection_type,
                                                          tuple(new_components))

  def _placeholder_value(self) -> Any:
    components = [
        component._placeholder_value()  # pylint: disable=protected-access
        for component in self.components
    ]
    return self.collection_type(*components)


class Attrs(NamedTuple):
  """Represents a class annotated by attr.s.

  Each attr.s class has a fixed, ordered set of attributes. Therefore, we only
  need to consider the class type and the underlying attributes. Extra
  metadata including attribute names can be ignored.
  """


class Dict(trace.TraceType):
  """Represents a dictionary of TraceType objects.

  Attributes:
    mapping: A mapping from keys to corresponding TraceTypes of the dict values.
  """

  def __init__(self, mapping: PythonDict[Hashable, trace.TraceType]):
    self.mapping = mapping

  def _has_same_structure(self, other):
    if not isinstance(other, Dict):
      return False

    return self.mapping.keys() == other.mapping.keys()

  def is_subtype_of(self, other: trace.TraceType) -> bool:
    """See base class."""
    if not self._has_same_structure(other):
      return False

    # We need all keys to be present because there can be logic relying on
    # their existence or lack thereof and hence can not guarantee subtype based
    # on a subset or superset of keys.
    # Only the tracing code can explicitly check for key dependencies and inform
    # that decision.
    return all(self.mapping[key].is_subtype_of(other.mapping[key])
               for key in self.mapping)

  def most_specific_common_supertype(
      self, types: Sequence[trace.TraceType]) -> Optional["Dict"]:
    """See base class."""
    if not all(self._has_same_structure(other) for other in types):
      return None

    new_mapping = {}
    for key in self.mapping.keys():
      common = self.mapping[key].most_specific_common_supertype(
          [other.mapping[key] for other in types])
      if common is None:
        return None
      else:
        new_mapping[key] = common

    return Dict(new_mapping)

  def _placeholder_value(self) -> Any:
    return {
        key: value._placeholder_value()  # pylint: disable=protected-access
        for key, value in self.mapping.items()
    }

  def __eq__(self, other) -> bool:
    if not isinstance(other, trace.TraceType):
      return NotImplemented

    if not isinstance(other, Dict):
      return False

    return self.mapping == other.mapping

  def __hash__(self) -> int:
    return hash(frozenset(self.mapping.keys()))

  def __repr__(self):
    return f"{self.__class__.__name__}(mapping={self.mapping!r})"


class Reference(trace.TraceType):
  """Represents a resource with an identifier.

  Resource identifiers are useful to denote identical resources, that is,
  resources which are known at compilation time to point to the same thing.
  This information is useful in automatic control dependencies for instance,
  where ops using the same resource don't run concurrently.
  """

  def __init__(self, base: trace.TraceType, identifier: Hashable):
    self.base = base
    self.identifier = identifier

  def is_subtype_of(self, other: trace.TraceType) -> bool:
    if isinstance(other, Reference) and self.identifier == other.identifier:
      return self.base.is_subtype_of(other.base)
    return False

  def most_specific_common_supertype(
      self, types: Sequence[trace.TraceType]) -> Optional["Reference"]:
    if all(
        isinstance(other, Reference) and self.identifier == other.identifier
        for other in types):
      base_supertype = self.base.most_specific_common_supertype(
          [other.base for other in types])
      if base_supertype is not None:
        return Reference(base_supertype, self.identifier)
    return None

  def _placeholder_value(self) -> Any:
    return self.base._placeholder_value()  # pylint: disable=protected-access

  def __eq__(self, other: Any) -> bool:
    if not isinstance(other, trace.TraceType):
      return NotImplemented

    return isinstance(
        other, Reference
    ) and self.identifier == other.identifier and self.base == other.base

  def __hash__(self) -> int:
    return hash((self.identifier, self.base))

  def __repr__(self):
    return (f"{self.__class__.__name__}(base={self.base!r}, "
            f"identifier={self.identifier!r})")



# Copyright 2015 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Helper functions for creating partitioned variables.

This is a convenient abstraction to partition a large variable across
multiple smaller variables that can be assigned to different devices.

The full variable can be reconstructed by concatenating the smaller variables.
Using partitioned variables instead of a single variable is mostly a
performance choice.  It however also has an impact on:

1. Random initialization, as the random number generator is called once per
   slice
2. Updates, as they happen in parallel across slices

A key design goal is to allow a different graph to repartition a variable
with the same name but different slicings, including possibly no partitions.

TODO(touts): If an initializer provides a seed, the seed must be changed
deterministically for each slice, maybe by adding one to it, otherwise each
slice will use the same values.  Maybe this can be done by passing the
slice offsets to the initializer functions.

Typical usage:

```python
# Create a list of partitioned variables with:
vs = create_partitioned_variables(
    <shape>, <slicing>, <initializer>, name=<optional-name>)

# Pass the list as inputs to embedding_lookup for sharded, parallel lookup:
y = embedding_lookup(vs, ids, partition_strategy="div")

# Or fetch the variables in parallel to speed up large matmuls:
z = matmul(x, concat(slice_dim, vs))
```
"""
import math

from tensorflow.python.framework import dtypes
from tensorflow.python.framework import tensor_shape
from tensorflow.python.ops import variable_scope
from tensorflow.python.util import deprecation
from tensorflow.python.util.tf_export import tf_export

__all__ = [
    "create_partitioned_variables",
    "variable_axis_size_partitioner",
    "min_max_variable_partitioner",
    "fixed_size_partitioner",
]


@tf_export(v1=["variable_axis_size_partitioner"])
def variable_axis_size_partitioner(
    max_shard_bytes, axis=0, bytes_per_string_element=16, max_shards=None):
  """Get a partitioner for VariableScope to keep shards below `max_shard_bytes`.

  This partitioner will shard a Variable along one axis, attempting to keep
  the maximum shard size below `max_shard_bytes`.  In practice, this is not
  always possible when sharding along only one axis.  When this happens,
  this axis is sharded as much as possible (i.e., every dimension becomes
  a separate shard).

  If the partitioner hits the `max_shards` limit, then each shard may end up
  larger than `max_shard_bytes`. By default `max_shards` equals `None` and no
  limit on the number of shards is enforced.

  One reasonable value for `max_shard_bytes` is `(64 << 20) - 1`, or almost
  `64MB`, to keep below the protobuf byte limit.

  Args:
    max_shard_bytes: The maximum size any given shard is allowed to be.
    axis: The axis to partition along.  Default: outermost axis.
    bytes_per_string_element: If the `Variable` is of type string, this provides
      an estimate of how large each scalar in the `Variable` is.
    max_shards: The maximum number of shards in int created taking precedence
      over `max_shard_bytes`.

  Returns:
    A partition function usable as the `partitioner` argument to
    `variable_scope` and `get_variable`.

  Raises:
    ValueError: If any of the byte counts are non-positive.
  """
  if max_shard_bytes < 1 or bytes_per_string_element < 1:
    raise ValueError(
        "Both max_shard_bytes and bytes_per_string_element must be positive. "
        f"Currently, max_shard_bytes is {max_shard_bytes} and"
        f"bytes_per_string_element is {bytes_per_string_element}")
  if max_shards and max_shards < 1:
    raise ValueError(
        "max_shards must be positive.")

  def _partitioner(shape, dtype):
    """Partitioner that partitions shards to have max_shard_bytes total size.

    Args:
      shape: A `TensorShape`.
      dtype: A `DType`.

    Returns:
      A tuple representing how much to slice each axis in shape.

    Raises:
      ValueError: If shape is not a fully defined `TensorShape` or dtype is not
        a `DType`.
    """
    if not isinstance(shape, tensor_shape.TensorShape):
      raise ValueError(f"shape is not a TensorShape: {shape}")
    if not shape.is_fully_defined():
      raise ValueError(f"shape is not fully defined: {shape}")
    if not isinstance(dtype, dtypes.DType):
      raise ValueError(f"dtype is not a DType: {dtype}")

    if dtype.base_dtype == dtypes.string:
      element_size = bytes_per_string_element
    else:
      element_size = dtype.size

    partitions = [1] * shape.ndims
    bytes_per_slice = 1.0 * (
        shape.num_elements() / shape.dims[axis].value) * element_size
    # How many slices can we fit on one shard of size at most max_shard_bytes?
    # At least one slice is required.
    slices_per_shard = max(1, math.floor(max_shard_bytes / bytes_per_slice))
    # How many shards do we need for axis given that each shard fits
    # slices_per_shard slices from a total of shape[axis] slices?
    axis_shards = int(math.ceil(
        1.0 * shape.dims[axis].value / slices_per_shard))
    if max_shards:
      axis_shards = min(max_shards, axis_shards)

    partitions[axis] = axis_shards

    return partitions

  return _partitioner


@tf_export(v1=["min_max_variable_partitioner"])
def min_max_variable_partitioner(max_partitions=1, axis=0,
                                 min_slice_size=256 << 10,
                                 bytes_per_string_element=16):
  """Partitioner to allocate minimum size per slice.

  Returns a partitioner that partitions the variable of given shape and dtype
  such that each partition has a minimum of `min_slice_size` slice of the
  variable. The maximum number of such partitions (upper bound) is given by
  `max_partitions`.

  Args:
    max_partitions: Upper bound on the number of partitions. Defaults to 1.
    axis: Axis along which to partition the variable. Defaults to 0.
    min_slice_size: Minimum size of the variable slice per partition. Defaults
      to 256K.
    bytes_per_string_element: If the `Variable` is of type string, this provides
      an estimate of how large each scalar in the `Variable` is.

  Returns:
    A partition function usable as the `partitioner` argument to
    `variable_scope` and `get_variable`.

  """
  def _partitioner(shape, dtype):
    """Partitioner that partitions list for a variable of given shape and type.

    Ex: Consider partitioning a variable of type float32 with
      shape=[1024, 1024].
      If `max_partitions` >= 16, this function would return
        [(1024 * 1024 * 4) / (256 * 1024), 1] = [16, 1].
      If `max_partitions` < 16, this function would return
        [`max_partitions`, 1].

    Args:
      shape: Shape of the variable.
      dtype: Type of the variable.

    Returns:
      List of partitions for each axis (currently only one axis can be
      partitioned).

    Raises:
      ValueError: If axis to partition along does not exist for the variable.
    """
    if axis >= len(shape):
      raise ValueError(
          f"Cannot partition variable along axis {axis} when shape is "
          f"only {shape}")
    if dtype.base_dtype == dtypes.string:
      bytes_per_element = bytes_per_string_element
    else:
      bytes_per_element = dtype.size
    total_size_bytes = shape.num_elements() * bytes_per_element
    partitions = total_size_bytes / min_slice_size
    partitions_list = [1] * len(shape)
    # We can not partition the variable beyond what its shape or
    # `max_partitions` allows.
    partitions_list[axis] = max(1, min(shape.dims[axis].value,
                                       max_partitions,
                                       int(math.ceil(partitions))))
    return partitions_list
  return _partitioner


@tf_export(v1=["fixed_size_partitioner"])
def fixed_size_partitioner(num_shards, axis=0):
  """Partitioner to specify a fixed number of shards along given axis.

  @compatibility(TF2)
  This API is deprecated in TF2. In TF2, partitioner is no longer part of
  the variable declaration via `tf.Variable`.
  [ParameterServer Training]
  (https://www.tensorflow.org/tutorials/distribute/parameter_server_training)
  handles partitioning of variables. The corresponding TF2 partitioner class of
  `fixed_size_partitioner` is
  `tf.distribute.experimental.partitioners.FixedShardsPartitioner`.

  Check the [migration guide]
  (https://www.tensorflow.org/guide/migrate#2_use_python_objects_to_track_variables_and_losses)
  on the differences in treatment of variables and losses between TF1 and TF2.

  Before:

    ```
    x = tf.compat.v1.get_variable(
      "x", shape=(2,), partitioner=tf.compat.v1.fixed_size_partitioner(2)
    )
    ```
  After:

    ```
    partitioner = (
        tf.distribute.experimental.partitioners.FixedShardsPartitioner(
            num_shards=2)
    )
    strategy = tf.distribute.experimental.ParameterServerStrategy(
                   cluster_resolver=cluster_resolver,
                   variable_partitioner=partitioner)

    with strategy.scope():
      x = tf.Variable([1.0, 2.0])
    ```
  @end_compatibility

  Args:
    num_shards: `int`, number of shards to partition variable.
    axis: `int`, axis to partition on.

  Returns:
    A partition function usable as the `partitioner` argument to
    `variable_scope` and `get_variable`.
  """
  def _partitioner(shape, **unused_args):
    partitions_list = [1] * len(shape)
    partitions_list[axis] = min(num_shards, shape.dims[axis].value)
    return partitions_list
  return _partitioner


@tf_export(v1=["create_partitioned_variables"])
@deprecation.deprecated(
    date=None,
    instructions="Use `tf.get_variable` with a partitioner set.")
def create_partitioned_variables(
    shape, slicing, initializer, dtype=dtypes.float32,
    trainable=True, collections=None, name=None, reuse=None):
  """Create a list of partitioned variables according to the given `slicing`.

  Currently only one dimension of the full variable can be sliced, and the
  full variable can be reconstructed by the concatenation of the returned
  list along that dimension.

  Args:
    shape: List of integers.  The shape of the full variable.
    slicing: List of integers.  How to partition the variable.
      Must be of the same length as `shape`.  Each value
      indicate how many slices to create in the corresponding
      dimension.  Presently only one of the values can be more than 1;
      that is, the variable can only be sliced along one dimension.

      For convenience, The requested number of partitions does not have to
      divide the corresponding dimension evenly.  If it does not, the
      shapes of the partitions are incremented by 1 starting from partition
      0 until all slack is absorbed.  The adjustment rules may change in the
      future, but as you can save/restore these variables with different
      slicing specifications this should not be a problem.
    initializer: A `Tensor` of shape `shape` or a variable initializer
      function.  If a function, it will be called once for each slice,
      passing the shape and data type of the slice as parameters.  The
      function must return a tensor with the same shape as the slice.
    dtype: Type of the variables. Ignored if `initializer` is a `Tensor`.
    trainable: If True also add all the variables to the graph collection
      `GraphKeys.TRAINABLE_VARIABLES`.
    collections: List of graph collections keys to add the variables to.
      Defaults to `[GraphKeys.GLOBAL_VARIABLES]`.
    name: Optional name for the full variable.  Defaults to
      `"PartitionedVariable"` and gets uniquified automatically.
    reuse: Boolean or `None`; if `True` and name is set, it would reuse
      previously created variables. if `False` it will create new variables.
      if `None`, it would inherit the parent scope reuse.

  Returns:
    A list of Variables corresponding to the slicing.

  Raises:
    ValueError: If any of the arguments is malformed.
  """
  if len(shape) != len(slicing):
    raise ValueError(
        "The 'shape' and 'slicing' of a partitioned Variable "
        f"must have the length: shape: {shape}, slicing: {slicing}")
  if len(shape) < 1:
    raise ValueError("A partitioned Variable must have rank at least 1: "
                     f"shape: {shape}")

  # Legacy: we are provided the slicing directly, so just pass it to
  # the partitioner.
  partitioner = lambda **unused_kwargs: slicing

  with variable_scope.variable_scope(
      name, "PartitionedVariable", reuse=reuse):
    # pylint: disable=protected-access
    partitioned_var = variable_scope._get_partitioned_variable(
        name=None,
        shape=shape,
        dtype=dtype,
        initializer=initializer,
        trainable=trainable,
        partitioner=partitioner,
        collections=collections)
    return list(partitioned_var)
    # pylint: enable=protected-access



# Copyright 2020 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""StructuredTensor array ops."""

from typing import Sequence

from tensorflow.python.framework import constant_op
from tensorflow.python.framework import dtypes
from tensorflow.python.framework import ops
from tensorflow.python.ops import array_ops
from tensorflow.python.ops import math_ops
from tensorflow.python.ops import random_ops
from tensorflow.python.ops.ragged import ragged_tensor
from tensorflow.python.ops.ragged.row_partition import RowPartition
from tensorflow.python.ops.structured.structured_tensor import StructuredTensor
from tensorflow.python.util import deprecation
from tensorflow.python.util import dispatch


@dispatch.dispatch_for_types(array_ops.expand_dims, StructuredTensor)
@deprecation.deprecated_args(None, 'Use the `axis` argument instead', 'dim')
def expand_dims(input, axis=None, name=None, dim=None):  # pylint: disable=redefined-builtin
  """Creates a StructuredTensor with a length 1 axis inserted at index `axis`.

  This is an implementation of tf.expand_dims for StructuredTensor. Note
  that the `axis` must be less than or equal to rank.

  >>> st = StructuredTensor.from_pyval([[{"x": 1}, {"x": 2}], [{"x": 3}]])
  >>> tf.expand_dims(st, 0).to_pyval()
  [[[{'x': 1}, {'x': 2}], [{'x': 3}]]]
  >>> tf.expand_dims(st, 1).to_pyval()
  [[[{'x': 1}, {'x': 2}]], [[{'x': 3}]]]
  >>> tf.expand_dims(st, 2).to_pyval()
  [[[{'x': 1}], [{'x': 2}]], [[{'x': 3}]]]
  >>> tf.expand_dims(st, -1).to_pyval()  # -1 is the same as 2
  [[[{'x': 1}], [{'x': 2}]], [[{'x': 3}]]]

  Args:
    input: the original StructuredTensor.
    axis: the axis to insert the dimension: `-(rank + 1) <= axis <= rank`
    name: the name of the op.
    dim: deprecated: use axis.

  Returns:
    a new structured tensor with larger rank.

  Raises:
    an error if `axis < -(rank + 1)` or `rank < axis`.
  """
  axis = deprecation.deprecated_argument_lookup('axis', axis, 'dim', dim)
  return _expand_dims_impl(input, axis, name=name)


@dispatch.dispatch_for_types(array_ops.expand_dims_v2, StructuredTensor)
def expand_dims_v2(input, axis, name=None):  # pylint: disable=redefined-builtin
  """Creates a StructuredTensor with a length 1 axis inserted at index `axis`.

  This is an implementation of tf.expand_dims for StructuredTensor. Note
  that the `axis` must be less than or equal to rank.

  >>> st = StructuredTensor.from_pyval([[{"x": 1}, {"x": 2}], [{"x": 3}]])
  >>> tf.expand_dims(st, 0).to_pyval()
  [[[{'x': 1}, {'x': 2}], [{'x': 3}]]]
  >>> tf.expand_dims(st, 1).to_pyval()
  [[[{'x': 1}, {'x': 2}]], [[{'x': 3}]]]
  >>> tf.expand_dims(st, 2).to_pyval()
  [[[{'x': 1}], [{'x': 2}]], [[{'x': 3}]]]
  >>> tf.expand_dims(st, -1).to_pyval()  # -1 is the same as 2
  [[[{'x': 1}], [{'x': 2}]], [[{'x': 3}]]]

  Args:
    input: the original StructuredTensor.
    axis: the axis to insert the dimension: `-(rank + 1) <= axis <= rank`
    name: the name of the op.

  Returns:
    a new structured tensor with larger rank.

  Raises:
    an error if `axis < -(rank + 1)` or `rank < axis`.
  """
  return _expand_dims_impl(input, axis, name=name)


@dispatch.dispatch_for_types(array_ops.gather, StructuredTensor)
def gather(params,
           indices,
           validate_indices=None,
           name=None,
           axis=None,
           batch_dims=0):
  """tf.gather for structured tensors.

  Does not support (yet) checks on illegal axis values, et cetera.

  Indices must be a ragged or dense tensor.
  Args:
    params: a structured tensor to be gathered
    indices: a ragged tensor or tensor to gather by.
    validate_indices: whether to validate the indices
    name: the name of the op(s).
    axis: the axis in params to gather on.
    batch_dims: the number of batch dimensions.

  Returns:
    the params reorganized according to indices.
  """
  if name is None:
    name = 'gather'
  with ops.name_scope(name):
    if axis is None:
      axis = batch_dims
    axis = array_ops.get_positive_axis(axis, params.shape.rank,
                                       ndims_name='params.shape.rank')
    indices = ragged_tensor.convert_to_tensor_or_ragged_tensor(
        indices, name='indices')

    def leaf_op(p):
      return array_ops.gather(
          p,
          indices,
          validate_indices=validate_indices,
          axis=axis,
          batch_dims=batch_dims,
          name=None)

    return _extend_op_single(params, leaf_op)


@dispatch.dispatch_for_types(array_ops.concat, StructuredTensor)
def concat(values, axis, name: str = 'concat'):
  """tf.concat for structured tensors.

  Does not support (yet) checks on illegal axis values, et cetera.

  Args:
    values: a sequence of StructuredTensors.
    axis: an axis to concatenate upon.
    name: the name of the op(s).

  Returns:
    the params reorganized according to indices.
  """
  if name is None:
    name = 'concat'
  _assert_concat_compatible_structured_tensors(values)
  def leaf_op(values):
    return array_ops.concat(values, axis)
  # TODO(martinz): handle axis when it is a tensor.
  axis = array_ops.get_positive_axis(axis, values[0].rank)
  with ops.name_scope(name, 'StructuredConcat', values):
    return _extend_op(values, leaf_op)


@dispatch.dispatch_for_types(random_ops.random_shuffle, StructuredTensor)
def random_shuffle(value, seed=None, name=None):
  """Shuffle a structured tensor on the zeroth axis.

  Args:
    value: a structured tensor of rank at least one.
    seed: the seed for shuffling.
    name: the name for shuffle.

  Returns:
    The shuffled structured tensor.
  """
  with ops.name_scope(name, 'shuffle', [value, seed]):
    if value.rank == 0:
      raise ValueError('Cannot shuffle a scalar StructuredTensor')
    first_dimension = value.nrows()
    index = random_ops.random_shuffle(math_ops.range(first_dimension),
                                      seed=seed)
    return gather(value, index, axis=0)


@dispatch.dispatch_for_types(array_ops.size_v2, StructuredTensor)
def size_v2(input, out_type=dtypes.int32, name=None):
  # pylint: disable=redefined-builtin
  """Returns the size of a tensor."""
  return size(input, name=name, out_type=out_type)


# pylint: disable=protected-access
@dispatch.dispatch_for_types(array_ops.size, StructuredTensor)
def size(input, name=None, out_type=dtypes.int32):
  # pylint: disable=redefined-builtin
  """Returns the size of a tensor."""
  with ops.name_scope(name, 'size', [input]) as name:
    if not input.row_partitions:
      if input.nrows() is not None:
        return math_ops.cast(input.nrows(), out_type)  # vector.
      else:
        return math_ops.cast(1, out_type)  # scalar.
    # 2D and up.
    nvals = input.row_partitions[-1].nvals()
    if nvals is None or out_type is None:
      return nvals
    return math_ops.cast(nvals, dtype=out_type)


# pylint: disable=protected-access
@dispatch.dispatch_for_types(array_ops.zeros_like, StructuredTensor)
def zeros_like(tensor, dtype=None, name=None, optimize=True):
  """Implementation of zeros_like for StructuredTensor for TF v1."""
  del optimize
  return zeros_like_v2(tensor, dtype=dtype, name=name)


# pylint: disable=protected-access
@dispatch.dispatch_for_types(array_ops.zeros_like_v2, StructuredTensor)
def zeros_like_v2(input, dtype=None, name=None):  # pylint: disable=redefined-builtin
  """Replace every object with a zero.

  Example:
  >>> st = StructuredTensor.from_pyval([{"x":[3]}, {"x":[4,5]}])
  >>> tf.zeros_like(st)
  <tf.Tensor: shape=(2,), dtype=int32, numpy=array([0.0, 0.0], dtype=float32)>
  >>> st = StructuredTensor.from_pyval([[{"x":[3]}], [{"x":[4,5]}, {"x":[]}]])
  >>> tf.zeros_like(st, dtype=tf.int32)
  <tf.RaggedTensor [[0], [0, 0]]>

  Args:
    input: a structured tensor.
    dtype: the dtype of the resulting zeros. (default is tf.float32)
    name: a name for the op.
  Returns:
    a tensor of zeros of the same shape.
  """
  if dtype is None:
    dtype = dtypes.float32
  with ops.name_scope(name, 'zeros_like', [input]) as name:
    if not input.row_partitions:
      if input.nrows() is not None:
        return array_ops.zeros([input.nrows()], dtype)  # vector.
      else:
        return array_ops.zeros([], dtype)  # scalar.
    # 2D and up.
    last_row_partition = input.row_partitions[-1]

    result = ragged_tensor.RaggedTensor._from_nested_row_partitions(
        array_ops.zeros(last_row_partition.nvals(), dtype=dtype),
        input.row_partitions)
    return result


# pylint: disable=protected-access
@dispatch.dispatch_for_types(array_ops.ones_like, StructuredTensor)
def ones_like(tensor, dtype=None, name=None, optimize=True):
  """Implementation of zeros_like for StructuredTensor for TF v1."""
  del optimize
  return ones_like_v2(tensor, dtype=dtype, name=name)


# pylint: disable=protected-access
@dispatch.dispatch_for_types(array_ops.ones_like_v2, StructuredTensor)
def ones_like_v2(input, dtype=None, name=None):  # pylint: disable=redefined-builtin
  """Replace every object with a zero.

  Example:
  >>> st = StructuredTensor.from_pyval([{"x":[3]}, {"x":[4,5]}])
  >>> tf.ones_like(st)
  <tf.Tensor: shape=(2,), dtype=int32, numpy=array([1.0, 1.0], dtype=float32)>
  >>> st = StructuredTensor.from_pyval([[{"x":[3]}], [{"x":[4,5]}, {"x":[]}]])
  >>> tf.ones_like(st, dtype=tf.int32)
  <tf.RaggedTensor [[1], [1, 1]]>

  Args:
    input: a structured tensor.
    dtype: the dtype of the resulting zeros. (default is tf.float32)
    name: a name for the op.
  Returns:
    a tensor of zeros of the same shape.
  """
  if dtype is None:
    dtype = dtypes.float32
  with ops.name_scope(name, 'ones_like', [input]) as name:
    if not input.row_partitions:
      if input.nrows() is not None:
        return array_ops.ones([input.nrows()], dtype)  # vector.
      else:
        return array_ops.ones([], dtype)  # scalar.
    # 2D and up.
    last_row_partition = input.row_partitions[-1]

    result = ragged_tensor.RaggedTensor._from_nested_row_partitions(
        array_ops.ones(last_row_partition.nvals(), dtype=dtype),
        input.row_partitions)
    return result


@dispatch.dispatch_for_types(array_ops.rank, StructuredTensor)
def rank(input, name=None):
  # pylint: disable=redefined-builtin
  """Returns the rank of a tensor."""
  with ops.name_scope(name, 'rank', [input]) as name:
    return constant_op.constant(input.rank, dtype=dtypes.int32)


def _expand_dims_impl(st, axis, name=None):  # pylint: disable=redefined-builtin
  """Creates a StructuredTensor with a length 1 axis inserted at index `axis`.

  This is an implementation of tf.expand_dims for StructuredTensor. Note
  that the `axis` must be less than or equal to rank.

  >>> st = StructuredTensor.from_pyval([[{"x": 1}, {"x": 2}], [{"x": 3}]])
  >>> tf.expand_dims(st, 0).to_pyval()
  [[[{'x': 1}, {'x': 2}], [{'x': 3}]]]
  >>> tf.expand_dims(st, 1).to_pyval()
  [[[{'x': 1}, {'x': 2}]], [[{'x': 3}]]]
  >>> tf.expand_dims(st, 2).to_pyval()
  [[[{'x': 1}], [{'x': 2}]], [[{'x': 3}]]]
  >>> tf.expand_dims(st, -1).to_pyval()  # -1 is the same as 2
  [[[{'x': 1}], [{'x': 2}]], [[{'x': 3}]]]

  Args:
    st: the original StructuredTensor.
    axis: the axis to insert the dimension: `-(rank + 1) <= axis <= rank`
    name: the name of the op.

  Returns:
    a new structured tensor with larger rank.

  Raises:
    an error if `axis < -(rank + 1)` or `rank < axis`.
  """
  axis = array_ops.get_positive_axis(
      axis, st.rank + 1, axis_name='axis', ndims_name='rank(st)')
  with ops.name_scope(name, 'ExpandDims', [st, axis]):
    new_fields = {
        k: array_ops.expand_dims(v, axis) for (k, v) in st._fields.items()
    }
    new_shape = st.shape[:axis] + (1,) + st.shape[axis:]
    new_row_partitions = _expand_st_row_partitions(st, axis)
    new_nrows = st.nrows() if (axis > 0) else 1
    return StructuredTensor.from_fields(
        new_fields,
        shape=new_shape,
        row_partitions=new_row_partitions,
        nrows=new_nrows)


def _expand_st_row_partitions(st, axis):
  """Create the row_partitions for expand_dims."""
  if axis == 0:
    if st.shape.rank == 0:
      return ()
    nvals = st.nrows()
    new_partition = RowPartition.from_uniform_row_length(
        nvals, nvals, nrows=1, validate=False)
    return (new_partition,) + st.row_partitions
  elif axis == st.rank:
    nvals = (
        st.row_partitions[axis - 2].nvals() if (axis - 2 >= 0) else st.nrows())
    return st.row_partitions + (RowPartition.from_uniform_row_length(
        1, nvals, nrows=nvals, validate=False),)
  else:
    nvals = (
        st.row_partitions[axis - 1].nrows() if (axis - 1 >= 0) else st.nrows())
    return st.row_partitions[:axis - 1] + (RowPartition.from_uniform_row_length(
        1, nvals, nrows=nvals, validate=False),) + st.row_partitions[axis - 1:]


# TODO(martinz): consider allowing values to be nested.
def _extend_op(values, leaf_op, empty_st_op=None):
  """Extend an op from RaggedTensor and Tensor to StructuredTensor.

  Visits all children of the structured tensor, and children of children,
  applying leaf_op whenever it reaches a leaf, and empty_st_op whenever
  it reaches an internal node without children.

  Args:
    values: a list of structured tensors, ragged tensors, or tensors. All must
      have the same type. If they are structured tensors, they must have the
      same paths.
    leaf_op: an op for handling non-structured tensor.
    empty_st_op: op to create a structured tensor without fields.

  Returns:
    the result of the extended op (a StructuredTensor, RaggedTensor, or Tensor)

  Raises:
    ValueError:
      If values is not a Sequence or is empty.
  """
  if not isinstance(values, Sequence):
    raise ValueError('Expected a list')

  if not values:
    raise ValueError('List cannot be empty')

  if empty_st_op is None:
    empty_st_op = empty_st_op_like_zeros(leaf_op)
  # Use the structure of the first StructuredTensor. They are all assumed to
  # be the same.
  value = values[0]

  if isinstance(value, StructuredTensor):
    # TODO(martinz): Calling empty_st_op may add unnecessary ops. Revisit later.
    empty_result = empty_st_op(values)
    if not value.field_names():
      return empty_result
    new_fields = {}
    for k in value.field_names():
      new_fields[k] = _extend_op([v.field_value(k) for v in values], leaf_op,
                                 empty_st_op)
    return StructuredTensor.from_fields(new_fields, shape=empty_result.shape)
  else:
    return leaf_op(values)


def _extend_op_single(value, leaf_op, empty_st_op=None):
  """Extend an op to a value instead of a list of values."""

  def to_list_op(element_op):
    if element_op is None:
      return None

    def list_op(values):
      [value] = values
      return element_op(value)

    return list_op

  return _extend_op([value], to_list_op(leaf_op), to_list_op(empty_st_op))


def empty_st_op_like_zeros(leaf_op):

  def empty_st_op(values):
    as_zeros = [
        zeros_like_v2(value, dtype=dtypes.int32) for value in values
    ]
    result = leaf_op(as_zeros)
    return _structured_tensor_like(result)

  return empty_st_op


def _structured_tensor_from_dense_tensor(t):
  """Create a structured tensor with the shape of a dense tensor."""
  # Note: If a tensor will have rank 0,
  # it either has a fully defined shape or has unknown rank.
  if t.shape.is_fully_defined():
    return StructuredTensor.from_fields({}, shape=t.shape)
  elif t.shape.rank is None:
    raise ValueError("Can't build StructuredTensor w/ unknown rank")
  elif t.shape.rank == 1:
    return StructuredTensor.from_fields({}, shape=t.shape,
                                        nrows=array_ops.shape(t)[0])
  else:
    rt = ragged_tensor.RaggedTensor.from_tensor(t)
    return _structured_tensor_from_row_partitions(t.shape,
                                                  rt._nested_row_partitions)


def _structured_tensor_from_row_partitions(shape, row_partitions):
  return StructuredTensor.from_fields({},
                                      shape=shape,
                                      row_partitions=row_partitions)


# pylint: disable=protected_access
def _all_nested_row_partitions(rt):
  """Returns all nested row partitions in rt, including for dense dimensions."""
  if isinstance(rt, ops.Tensor):
    if rt.shape.rank <= 1:
      return ()
    else:
      rt2 = ragged_tensor.RaggedTensor.from_tensor(rt)
      return rt2._nested_row_partitions
  else:
    tail_partitions = _all_nested_row_partitions(rt.flat_values)
    head_partitions = rt._nested_row_partitions  # pylint: disable=protected_access
    return head_partitions + tail_partitions


def _structured_tensor_like(t):
  """Create a StructuredTensor with the shape of a (composite) tensor."""
  if isinstance(t, ops.Tensor):
    return _structured_tensor_from_dense_tensor(t)
  if ragged_tensor.is_ragged(t):
    return StructuredTensor.from_fields(
        {}, shape=t.get_shape(), row_partitions=_all_nested_row_partitions(t))
  # here, it is a StructuredTensor
  return StructuredTensor.from_fields({},
                                      shape=t.shape,
                                      row_partitions=t.row_partitions,
                                      nrows=t.nrows())


def _get_all_paths(st):
  """Get all the paths from a StructuredTensor."""
  fields = st.field_names()
  all_paths = {()}
  for k in fields:
    v = st.field_value(k)
    if isinstance(v, StructuredTensor):
      all_paths = all_paths.union([(k,) + p for p in _get_all_paths(v)])
    else:
      all_paths.add((k,))
  return all_paths


def _get_all_ranks(st):
  """Get ranks of all submessages of a StructuredTensor."""
  fields = st.field_names()
  all_ranks = {(): st.rank}
  for k in fields:
    v = st.field_value(k)
    if isinstance(v, StructuredTensor):
      for (k2, v2) in _get_all_ranks(v).items():
        all_ranks[(k,) + k2] = v2
  return all_ranks


def _assert_all_paths_match(values):
  """Raises an error if the paths are not identical."""
  paths = [_get_all_paths(st) for st in values]
  path_diff = set()
  for other_paths in paths[1:]:
    path_diff = path_diff.union(paths[0].symmetric_difference(other_paths))
  if path_diff:
    raise ValueError(
        'Some paths are present in some, but not all, structured tensors: %r' %
        (path_diff,))


def _assert_all_ranks_match(values):
  """Raises an error if the ranks of submessages are not identical."""
  ranks = [_get_all_ranks(st) for st in values]
  for other_ranks in ranks[1:]:
    if other_ranks != ranks[0]:
      # TODO(martinz): If this becomes common, we can provide more detail.
      # e.g.: which path is inconsistent.
      raise ValueError('Ranks of sub-message do not match')


def _assert_concat_compatible_structured_tensors(values):
  """Sometimes raises an error if concat doesn't make sense statically on values.

  values must be a sequence, and each element in values must be a structured
  tensor, and must have the same paths. Additionally, each path that is a
  submessage must have the same rank.

  These constraints are sufficient for concat on the fields to be the same
  as concat on structured tensors. This is meant to capture scenarios like
  paths that are not in the first structured tensor, but are in later
  structured tensors, which will just be ignored by the recursive algorithm.

  If the rank of a submessage was different for two structured tensors,
  then that is also a non-sensical merge.

  Note that all of these checks are static, as paths and submessage ranks
  are known.

  Args:
    values: a Sequence of StructuredTensors.

  Raises:
    ValueError: if there is any inconsistency as described above.
  """
  if not isinstance(values, Sequence):
    raise ValueError('values must be a list of StructuredTensors (not a list)')
  if not values:
    raise ValueError('values must not be an empty list')
  for st in values:
    if not isinstance(st, StructuredTensor):
      raise ValueError('values must be a list of StructuredTensors')
  _assert_all_paths_match(values)
  _assert_all_ranks_match(values)



# This file is MACHINE GENERATED! Do not edit.
# Generated by: tensorflow/python/tools/api/generator/create_python_api.py script.
"""# tf.experimental.numpy: NumPy API on TensorFlow.

This module provides a subset of NumPy API, built on top of TensorFlow
operations. APIs are based on and have been tested with NumPy 1.16 version.

The set of supported APIs may be expanded over time. Also future releases may
change the baseline version of NumPy API being supported. A list of some
systematic differences with NumPy is listed later in the "Differences with
NumPy" section.

## Getting Started

Please also see [TensorFlow NumPy Guide](
https://www.tensorflow.org/guide/tf_numpy).

In the code snippets below, we will assume that `tf.experimental.numpy` is
imported as `tnp` and NumPy is imported as `np`

```python
print(tnp.ones([2,1]) + np.ones([1, 2]))
```

## Types

The module provides an `ndarray` class which wraps an immutable `tf.Tensor`.
Additional functions are provided which accept array-like objects. Here
array-like objects include `ndarrays` as defined by this module, as well as
`tf.Tensor`, in addition to types accepted by NumPy.

A subset of NumPy dtypes are supported. Type promotion follows NumPy
semantics.

```python
print(tnp.ones([1, 2], dtype=tnp.int16) + tnp.ones([2, 1], dtype=tnp.uint8))
```

## Array Interface

The `ndarray` class implements the `__array__` interface. This should allow
these objects to be passed into contexts that expect a NumPy or array-like
object (e.g. matplotlib).

```python
np.sum(tnp.ones([1, 2]) + np.ones([2, 1]))
```


## TF Interoperability

The TF-NumPy API calls can be interleaved with TensorFlow calls
without incurring Tensor data copies. This is true even if the `ndarray` or
`tf.Tensor` is placed on a non-CPU device.

In general, the expected behavior should be on par with that of code involving
`tf.Tensor` and running stateless TensorFlow functions on them.

```python
tnp.sum(tnp.ones([1, 2]) + tf.ones([2, 1]))
```

Note that the `__array_priority__` is currently chosen to be lower than
`tf.Tensor`. Hence the `+` operator above returns a `tf.Tensor`.

Additional examples of interoperability include:

*  using `with tf.GradientTape()` scope to compute gradients through the
  TF-NumPy API calls.
*  using `tf.distribution.Strategy` scope for distributed execution
*  using `tf.vectorized_map()` for speeding up code using auto-vectorization



## Device Support

Given that `ndarray` and functions wrap TensorFlow constructs, the code will
have GPU and TPU support on par with TensorFlow. Device placement can be
controlled by using `with tf.device` scopes. Note that these devices could
be local or remote.

```python
with tf.device("GPU:0"):
  x = tnp.ones([1, 2])
print(tf.convert_to_tensor(x).device)
```

## Graph and Eager Modes

Eager mode execution should typically match NumPy semantics of executing
op-by-op. However the same code can be executed in graph mode, by putting it
inside a `tf.function`. The function body can contain NumPy code, and the inputs
can be `ndarray` as well.

```python
@tf.function
def f(x, y):
  return tnp.sum(x + y)

f(tnp.ones([1, 2]), tf.ones([2, 1]))
```
Python control flow based on `ndarray` values will be translated by
[autograph](https://www.tensorflow.org/code/tensorflow/python/autograph/g3doc/reference/index.md)
into `tf.cond` and `tf.while_loop` constructs. The code can be XLA compiled
for further optimizations.

However, note that graph mode execution can change behavior of certain
operations since symbolic execution may not have information that is computed
during runtime. Some differences are:

*   Shapes can be incomplete or unknown in graph mode. This means that
    `ndarray.shape`, `ndarray.size` and `ndarray.ndim` can return `ndarray`
    objects instead of returning integer (or tuple of integer) values.
*   `__len__`, `__iter__` and `__index__` properties of `ndarray`
    may similarly not be supported in graph mode. Code using these
    may need to change to explicit shape operations or control flow
    constructs.
*   Also note the [autograph limitations](
https://github.com/tensorflow/tensorflow/blob/master/tensorflow/python/autograph/g3doc/reference/limitations.md).


## Mutation and Variables

`ndarrays` currently wrap immutable `tf.Tensor`. Hence mutation
operations like slice assigns are not supported. This may change in the future.
Note however that one can directly construct a `tf.Variable` and use that with
the TF-NumPy APIs.

```python
tf_var = tf.Variable(2.0)
tf_var.assign_add(tnp.square(tf_var))
```

## Differences with NumPy

Here is a non-exhaustive list of differences:

*   Not all dtypes are currently supported. e.g. `np.float96`, `np.float128`.
    `np.object_`, `np.str_`, `np.recarray` types are not supported.
*   `ndarray` storage is in C order only. Fortran order, views, `stride_tricks`
    are not supported.
*   Only a subset of functions and modules are supported. This set will be
    expanded over time. For supported functions, some arguments or argument
    values may not be supported. These differences are generally provided in the
    function comments. Full `ufunc` support is also not provided.
*   Buffer mutation is currently not supported. `ndarrays` wrap immutable
    tensors. This means that output buffer arguments (e.g. `out` in ufuncs) are
    not supported.
*   NumPy C API is not supported. NumPy's Cython and Swig integration are not
    supported.

"""

import sys as _sys

from . import random
from tensorflow.python.framework.ops import Tensor as ndarray
from tensorflow.python.ops.numpy_ops import issubdtype
from tensorflow.python.ops.numpy_ops import max
from tensorflow.python.ops.numpy_ops import min
from tensorflow.python.ops.numpy_ops import round
from tensorflow.python.ops.numpy_ops.np_array_ops import all
from tensorflow.python.ops.numpy_ops.np_array_ops import amax
from tensorflow.python.ops.numpy_ops.np_array_ops import amin
from tensorflow.python.ops.numpy_ops.np_array_ops import any
from tensorflow.python.ops.numpy_ops.np_array_ops import arange
from tensorflow.python.ops.numpy_ops.np_array_ops import around
from tensorflow.python.ops.numpy_ops.np_array_ops import array
from tensorflow.python.ops.numpy_ops.np_array_ops import asanyarray
from tensorflow.python.ops.numpy_ops.np_array_ops import asarray
from tensorflow.python.ops.numpy_ops.np_array_ops import ascontiguousarray
from tensorflow.python.ops.numpy_ops.np_array_ops import atleast_1d
from tensorflow.python.ops.numpy_ops.np_array_ops import atleast_2d
from tensorflow.python.ops.numpy_ops.np_array_ops import atleast_3d
from tensorflow.python.ops.numpy_ops.np_array_ops import broadcast_arrays
from tensorflow.python.ops.numpy_ops.np_array_ops import broadcast_to
from tensorflow.python.ops.numpy_ops.np_array_ops import compress
from tensorflow.python.ops.numpy_ops.np_array_ops import copy
from tensorflow.python.ops.numpy_ops.np_array_ops import cumprod
from tensorflow.python.ops.numpy_ops.np_array_ops import cumsum
from tensorflow.python.ops.numpy_ops.np_array_ops import diag
from tensorflow.python.ops.numpy_ops.np_array_ops import diag_indices
from tensorflow.python.ops.numpy_ops.np_array_ops import diagflat
from tensorflow.python.ops.numpy_ops.np_array_ops import diagonal
from tensorflow.python.ops.numpy_ops.np_array_ops import dsplit
from tensorflow.python.ops.numpy_ops.np_array_ops import dstack
from tensorflow.python.ops.numpy_ops.np_array_ops import empty
from tensorflow.python.ops.numpy_ops.np_array_ops import empty_like
from tensorflow.python.ops.numpy_ops.np_array_ops import expand_dims
from tensorflow.python.ops.numpy_ops.np_array_ops import eye
from tensorflow.python.ops.numpy_ops.np_array_ops import flip
from tensorflow.python.ops.numpy_ops.np_array_ops import fliplr
from tensorflow.python.ops.numpy_ops.np_array_ops import flipud
from tensorflow.python.ops.numpy_ops.np_array_ops import full
from tensorflow.python.ops.numpy_ops.np_array_ops import full_like
from tensorflow.python.ops.numpy_ops.np_array_ops import hsplit
from tensorflow.python.ops.numpy_ops.np_array_ops import hstack
from tensorflow.python.ops.numpy_ops.np_array_ops import identity
from tensorflow.python.ops.numpy_ops.np_array_ops import imag
from tensorflow.python.ops.numpy_ops.np_array_ops import isscalar
from tensorflow.python.ops.numpy_ops.np_array_ops import ix_
from tensorflow.python.ops.numpy_ops.np_array_ops import mean
from tensorflow.python.ops.numpy_ops.np_array_ops import moveaxis
from tensorflow.python.ops.numpy_ops.np_array_ops import ndim
from tensorflow.python.ops.numpy_ops.np_array_ops import newaxis
from tensorflow.python.ops.numpy_ops.np_array_ops import nonzero
from tensorflow.python.ops.numpy_ops.np_array_ops import ones
from tensorflow.python.ops.numpy_ops.np_array_ops import ones_like
from tensorflow.python.ops.numpy_ops.np_array_ops import pad
from tensorflow.python.ops.numpy_ops.np_array_ops import prod
from tensorflow.python.ops.numpy_ops.np_array_ops import ravel
from tensorflow.python.ops.numpy_ops.np_array_ops import real
from tensorflow.python.ops.numpy_ops.np_array_ops import repeat
from tensorflow.python.ops.numpy_ops.np_array_ops import reshape
from tensorflow.python.ops.numpy_ops.np_array_ops import roll
from tensorflow.python.ops.numpy_ops.np_array_ops import rot90
from tensorflow.python.ops.numpy_ops.np_array_ops import select
from tensorflow.python.ops.numpy_ops.np_array_ops import shape
from tensorflow.python.ops.numpy_ops.np_array_ops import sign
from tensorflow.python.ops.numpy_ops.np_array_ops import size
from tensorflow.python.ops.numpy_ops.np_array_ops import split
from tensorflow.python.ops.numpy_ops.np_array_ops import squeeze
from tensorflow.python.ops.numpy_ops.np_array_ops import stack
from tensorflow.python.ops.numpy_ops.np_array_ops import std
from tensorflow.python.ops.numpy_ops.np_array_ops import sum
from tensorflow.python.ops.numpy_ops.np_array_ops import swapaxes
from tensorflow.python.ops.numpy_ops.np_array_ops import take
from tensorflow.python.ops.numpy_ops.np_array_ops import take_along_axis
from tensorflow.python.ops.numpy_ops.np_array_ops import transpose
from tensorflow.python.ops.numpy_ops.np_array_ops import tri
from tensorflow.python.ops.numpy_ops.np_array_ops import tril
from tensorflow.python.ops.numpy_ops.np_array_ops import triu
from tensorflow.python.ops.numpy_ops.np_array_ops import vander
from tensorflow.python.ops.numpy_ops.np_array_ops import var
from tensorflow.python.ops.numpy_ops.np_array_ops import vsplit
from tensorflow.python.ops.numpy_ops.np_array_ops import vstack
from tensorflow.python.ops.numpy_ops.np_array_ops import where
from tensorflow.python.ops.numpy_ops.np_array_ops import zeros
from tensorflow.python.ops.numpy_ops.np_array_ops import zeros_like
from tensorflow.python.ops.numpy_ops.np_config import enable_numpy_behavior as experimental_enable_numpy_behavior
from tensorflow.python.ops.numpy_ops.np_dtypes import bool_
from tensorflow.python.ops.numpy_ops.np_dtypes import complex128
from tensorflow.python.ops.numpy_ops.np_dtypes import complex64
from tensorflow.python.ops.numpy_ops.np_dtypes import complex_
from tensorflow.python.ops.numpy_ops.np_dtypes import float16
from tensorflow.python.ops.numpy_ops.np_dtypes import float32
from tensorflow.python.ops.numpy_ops.np_dtypes import float64
from tensorflow.python.ops.numpy_ops.np_dtypes import float_
from tensorflow.python.ops.numpy_ops.np_dtypes import iinfo
from tensorflow.python.ops.numpy_ops.np_dtypes import inexact
from tensorflow.python.ops.numpy_ops.np_dtypes import int16
from tensorflow.python.ops.numpy_ops.np_dtypes import int32
from tensorflow.python.ops.numpy_ops.np_dtypes import int64
from tensorflow.python.ops.numpy_ops.np_dtypes import int8
from tensorflow.python.ops.numpy_ops.np_dtypes import int_
from tensorflow.python.ops.numpy_ops.np_dtypes import object_
from tensorflow.python.ops.numpy_ops.np_dtypes import string_
from tensorflow.python.ops.numpy_ops.np_dtypes import uint16
from tensorflow.python.ops.numpy_ops.np_dtypes import uint32
from tensorflow.python.ops.numpy_ops.np_dtypes import uint64
from tensorflow.python.ops.numpy_ops.np_dtypes import uint8
from tensorflow.python.ops.numpy_ops.np_dtypes import unicode_
from tensorflow.python.ops.numpy_ops.np_math_ops import abs
from tensorflow.python.ops.numpy_ops.np_math_ops import absolute
from tensorflow.python.ops.numpy_ops.np_math_ops import add
from tensorflow.python.ops.numpy_ops.np_math_ops import allclose
from tensorflow.python.ops.numpy_ops.np_math_ops import angle
from tensorflow.python.ops.numpy_ops.np_math_ops import append
from tensorflow.python.ops.numpy_ops.np_math_ops import arccos
from tensorflow.python.ops.numpy_ops.np_math_ops import arccosh
from tensorflow.python.ops.numpy_ops.np_math_ops import arcsin
from tensorflow.python.ops.numpy_ops.np_math_ops import arcsinh
from tensorflow.python.ops.numpy_ops.np_math_ops import arctan
from tensorflow.python.ops.numpy_ops.np_math_ops import arctan2
from tensorflow.python.ops.numpy_ops.np_math_ops import arctanh
from tensorflow.python.ops.numpy_ops.np_math_ops import argmax
from tensorflow.python.ops.numpy_ops.np_math_ops import argmin
from tensorflow.python.ops.numpy_ops.np_math_ops import argsort
from tensorflow.python.ops.numpy_ops.np_math_ops import array_equal
from tensorflow.python.ops.numpy_ops.np_math_ops import average
from tensorflow.python.ops.numpy_ops.np_math_ops import bitwise_and
from tensorflow.python.ops.numpy_ops.np_math_ops import bitwise_not
from tensorflow.python.ops.numpy_ops.np_math_ops import bitwise_or
from tensorflow.python.ops.numpy_ops.np_math_ops import bitwise_xor
from tensorflow.python.ops.numpy_ops.np_math_ops import cbrt
from tensorflow.python.ops.numpy_ops.np_math_ops import ceil
from tensorflow.python.ops.numpy_ops.np_math_ops import clip
from tensorflow.python.ops.numpy_ops.np_math_ops import concatenate
from tensorflow.python.ops.numpy_ops.np_math_ops import conj
from tensorflow.python.ops.numpy_ops.np_math_ops import conjugate
from tensorflow.python.ops.numpy_ops.np_math_ops import cos
from tensorflow.python.ops.numpy_ops.np_math_ops import cosh
from tensorflow.python.ops.numpy_ops.np_math_ops import count_nonzero
from tensorflow.python.ops.numpy_ops.np_math_ops import cross
from tensorflow.python.ops.numpy_ops.np_math_ops import deg2rad
from tensorflow.python.ops.numpy_ops.np_math_ops import diff
from tensorflow.python.ops.numpy_ops.np_math_ops import divide
from tensorflow.python.ops.numpy_ops.np_math_ops import divmod
from tensorflow.python.ops.numpy_ops.np_math_ops import dot
from tensorflow.python.ops.numpy_ops.np_math_ops import e
from tensorflow.python.ops.numpy_ops.np_math_ops import einsum
from tensorflow.python.ops.numpy_ops.np_math_ops import equal
from tensorflow.python.ops.numpy_ops.np_math_ops import exp
from tensorflow.python.ops.numpy_ops.np_math_ops import exp2
from tensorflow.python.ops.numpy_ops.np_math_ops import expm1
from tensorflow.python.ops.numpy_ops.np_math_ops import fabs
from tensorflow.python.ops.numpy_ops.np_math_ops import fix
from tensorflow.python.ops.numpy_ops.np_math_ops import float_power
from tensorflow.python.ops.numpy_ops.np_math_ops import floor
from tensorflow.python.ops.numpy_ops.np_math_ops import floor_divide
from tensorflow.python.ops.numpy_ops.np_math_ops import gcd
from tensorflow.python.ops.numpy_ops.np_math_ops import geomspace
from tensorflow.python.ops.numpy_ops.np_math_ops import greater
from tensorflow.python.ops.numpy_ops.np_math_ops import greater_equal
from tensorflow.python.ops.numpy_ops.np_math_ops import heaviside
from tensorflow.python.ops.numpy_ops.np_math_ops import hypot
from tensorflow.python.ops.numpy_ops.np_math_ops import inf
from tensorflow.python.ops.numpy_ops.np_math_ops import inner
from tensorflow.python.ops.numpy_ops.np_math_ops import isclose
from tensorflow.python.ops.numpy_ops.np_math_ops import iscomplex
from tensorflow.python.ops.numpy_ops.np_math_ops import iscomplexobj
from tensorflow.python.ops.numpy_ops.np_math_ops import isfinite
from tensorflow.python.ops.numpy_ops.np_math_ops import isinf
from tensorflow.python.ops.numpy_ops.np_math_ops import isnan
from tensorflow.python.ops.numpy_ops.np_math_ops import isneginf
from tensorflow.python.ops.numpy_ops.np_math_ops import isposinf
from tensorflow.python.ops.numpy_ops.np_math_ops import isreal
from tensorflow.python.ops.numpy_ops.np_math_ops import isrealobj
from tensorflow.python.ops.numpy_ops.np_math_ops import kron
from tensorflow.python.ops.numpy_ops.np_math_ops import lcm
from tensorflow.python.ops.numpy_ops.np_math_ops import less
from tensorflow.python.ops.numpy_ops.np_math_ops import less_equal
from tensorflow.python.ops.numpy_ops.np_math_ops import linspace
from tensorflow.python.ops.numpy_ops.np_math_ops import log
from tensorflow.python.ops.numpy_ops.np_math_ops import log10
from tensorflow.python.ops.numpy_ops.np_math_ops import log1p
from tensorflow.python.ops.numpy_ops.np_math_ops import log2
from tensorflow.python.ops.numpy_ops.np_math_ops import logaddexp
from tensorflow.python.ops.numpy_ops.np_math_ops import logaddexp2
from tensorflow.python.ops.numpy_ops.np_math_ops import logical_and
from tensorflow.python.ops.numpy_ops.np_math_ops import logical_not
from tensorflow.python.ops.numpy_ops.np_math_ops import logical_or
from tensorflow.python.ops.numpy_ops.np_math_ops import logical_xor
from tensorflow.python.ops.numpy_ops.np_math_ops import logspace
from tensorflow.python.ops.numpy_ops.np_math_ops import matmul
from tensorflow.python.ops.numpy_ops.np_math_ops import maximum
from tensorflow.python.ops.numpy_ops.np_math_ops import meshgrid
from tensorflow.python.ops.numpy_ops.np_math_ops import minimum
from tensorflow.python.ops.numpy_ops.np_math_ops import mod
from tensorflow.python.ops.numpy_ops.np_math_ops import multiply
from tensorflow.python.ops.numpy_ops.np_math_ops import nanmean
from tensorflow.python.ops.numpy_ops.np_math_ops import nanprod
from tensorflow.python.ops.numpy_ops.np_math_ops import nansum
from tensorflow.python.ops.numpy_ops.np_math_ops import negative
from tensorflow.python.ops.numpy_ops.np_math_ops import nextafter
from tensorflow.python.ops.numpy_ops.np_math_ops import not_equal
from tensorflow.python.ops.numpy_ops.np_math_ops import outer
from tensorflow.python.ops.numpy_ops.np_math_ops import pi
from tensorflow.python.ops.numpy_ops.np_math_ops import polyval
from tensorflow.python.ops.numpy_ops.np_math_ops import positive
from tensorflow.python.ops.numpy_ops.np_math_ops import power
from tensorflow.python.ops.numpy_ops.np_math_ops import ptp
from tensorflow.python.ops.numpy_ops.np_math_ops import rad2deg
from tensorflow.python.ops.numpy_ops.np_math_ops import reciprocal
from tensorflow.python.ops.numpy_ops.np_math_ops import remainder
from tensorflow.python.ops.numpy_ops.np_math_ops import signbit
from tensorflow.python.ops.numpy_ops.np_math_ops import sin
from tensorflow.python.ops.numpy_ops.np_math_ops import sinc
from tensorflow.python.ops.numpy_ops.np_math_ops import sinh
from tensorflow.python.ops.numpy_ops.np_math_ops import sort
from tensorflow.python.ops.numpy_ops.np_math_ops import sqrt
from tensorflow.python.ops.numpy_ops.np_math_ops import square
from tensorflow.python.ops.numpy_ops.np_math_ops import subtract
from tensorflow.python.ops.numpy_ops.np_math_ops import tan
from tensorflow.python.ops.numpy_ops.np_math_ops import tanh
from tensorflow.python.ops.numpy_ops.np_math_ops import tensordot
from tensorflow.python.ops.numpy_ops.np_math_ops import tile
from tensorflow.python.ops.numpy_ops.np_math_ops import trace
from tensorflow.python.ops.numpy_ops.np_math_ops import true_divide
from tensorflow.python.ops.numpy_ops.np_math_ops import vdot
from tensorflow.python.ops.numpy_ops.np_utils import finfo
from tensorflow.python.ops.numpy_ops.np_utils import promote_types
from tensorflow.python.ops.numpy_ops.np_utils import result_type